#!/usr/bin/perl -w

#$Header: /home2/cvsroot/LogTrend/Agent/SnortAgent.pm,v 1.6 2001/11/20 16:28:30 lsimonneau Exp $
##******************************************************************************
## Class SnortAgent  isa  Agent
##  Description  : class for Snort supervision
##  Project      : LogTrend 1.0.0.0 - Atrid Systemes
##  Author       : Franois Dsarmnien f.desarmenien@atrid.fr
##  Author       : David Mascle d.mascle@atrid.fr
##******************************************************************************
#$Log: SnortAgent.pm,v $
#Revision 1.6  2001/11/20 16:28:30  lsimonneau
#Change use Getopt::Long for more compatibility with standard debian potato packages.
#
#Revision 1.5  2001/09/17 14:38:49  fdesar
#Typo in doc
#
#Revision 1.4  2001/08/05 13:42:01  fdesar
#use sigtrap to catch "normal" termination signals, so the snort daemon is
#aborted when the agent is killed.
#
#Revision 1.3  2001/07/27 15:50:20  fdesar
#A new LogTrend agent is born in CVS: SNMPAgent !!!
#
#Revision 1.2  2001/07/20 14:03:51  fdesar
#Modify snort STDERR following so it
#works with later version of snort
#
#Revision 1.1  2001/07/19 16:40:27  fdesar
#
#Moved module files to the right directories
#Updated package names and uses to reflect those changes
#Corrected bug in SnortAgent.pm for negating first value in SID parsing
#
#Revision 1.10  2001/07/18 16:29:35  fdesar
#
#Filtrage par SID
#Langage de specification de filtres de SID
#Tests << in vitro >>
#Documentation mise a jour
#A faire : tests << in vivo >>
#
#Revision 1.9  2001/06/20 17:27:50  fdesar
#
#Option 'order' ajoute aux regles
#MAJ doc pour option 'order'
#
#Revision 1.8  2001/06/19 16:38:40  fdesar
#
#More POD...
#
#Revision 1.7  2001/06/18 16:42:19  fdesar
#
#Documentation au format POD
#
#Revision 1.6  2001/06/11 07:46:05  fdesar
#
#Minor changes
#
#Revision 1.5  2001/06/07 15:17:51  fdesar
#
#Corected bug in Alarm collection
#
#Revision 1.4  2001/06/07 12:52:05  fdesar
#
#Added renaming/archiving of alert files
#
#Revision 1.3  2001/06/06 17:25:13  fdesar
#
#Restructuration of config file
#Added classifications in config file
#Added filters for preprocessors
#Rewrite of alert and alarms processing
#Added snort death alarm
#Need testing...
#
#Revision 1.2  2001/05/31 18:18:00  fdesar
#
#SnortAgent.pm is now roughly completed. Need testing & cleanups.
#
#Revision 1.1  2001/05/29 13:55:43  fdesar
#*** empty log message ***
#
#

package LogTrend::Agent::SnortAgent;

use strict;

use vars qw( @ISA );
use sigtrap qw(die INT TERM QUIT HUP);

use IO::Pipe;
use XML::DOM;
use LogTrend::Agent;
use LogTrend::Common::LogDie;
use Getopt::Long;

@ISA = ("LogTrend::Agent");

my $name = "SnortAgent";
my $version = "1.0.0.0";

my $death = 0;
my $snortpid = 0;

END {
        $snortpid
    and do {
        kill 9, $snortpid;
        SysLog("snort[$snortpid] terminated");
    }
}

##******************************************************************************
## Constructor  public > Agent
##  Description  : creat a new SnortAgent
##  Parameters   : none
##******************************************************************************
sub new
{
    my ($class) = shift;
    my($file) = @_;

    $class = ref($class) || $class;
    my $self = $class->SUPER::new( $name, $version );

    bless($self, $class);


    $self
}

##******************************************************************************
## Method ParseAgentOptions  public  (>Agent)
##  Description  : parses agent specific switches (overload to enable)
##  Parameters   : none
##  Return value : none
##******************************************************************************
#sub ParseAgentOptions {
#    my($self) = shift;
#    my($opts) = {};
#
#        GetOptions($opts, 'foo|f', 'bar|b=s', 'zork|z=i')
#    or  do {
#        print "Usage: blah, blah...";
#        exit(1);
#    };
#    $self->{Options} = $opts;
#}

##******************************************************************************
## Method ParseXMLConfigFile  public  (>Agent)
##  Description  : parses the XML config file
##  Parameters   : the file name to parse, the agent state
##  Return value : none
##******************************************************************************
sub ParseXMLConfigFile
{
   my ($self,$file,$agentstate) = @_;

   $self->SUPER::ParseXMLConfigFile( $file, $agentstate );

   # Parse Specific Config
   $self->_ParseConfiguration($file);

   # Run the Snort process
       $agentstate == 2
   and $self->_SpawnSnort();

}

##******************************************************************************
## Method CreateAgentDescription  public  (>Agent)
##  Description  : creates an agent's description
##  Parameters   : none
##  Return value : none
##******************************************************************************
sub CreateAgentDescription
{
   my $self = shift;
   my $dnum = 0;
   my $anum = 0;
   my($agconf) = $self->{SnortAgent};
   my($tags) = $agconf->{Classification}{Tag};

   for my $key (sort { $tags->{$a} <=> $tags->{$b} } keys(%$tags)) {
       $self->AddADataDescription(++$dnum, "Integer", "none", $key, "");
   }
       exists($agconf->{Alarms})
   and do {
       my($alist) = $agconf->{Alarms}{List};
       for my $alarm (@$alist) {
           my($label, $type) = @$alarm;
           $self->AddAnAlarmDescription(++$anum, $type, $label, "" );
       }
   };
}

##******************************************************************************
## Method CollectData  public  (>Agent)
##  Description  : collects data and alarms
##  Parameters   : none
##  Return value : none
##******************************************************************************
sub CollectData
{
   my $self = shift;
   my($collection, $alarms) = $self->_Collection();
   my $dnum = 0;
   my $anum = 0;

   for my $data (@$collection) {
       ++$dnum;
           $data
       and $self->AddDataInteger($dnum, $data );
   }

   for my $alarm (@$alarms) {
       ++$anum;
           $alarm
       and $self->AddAlarm( $anum );
   }

        $death
    and do {
        my($pid)=wait();
        my($agconf) = $self->{SnortAgent};

        $self->SendDie("snort[$pid] on interface '$agconf->{Interface}' " .
                       "died unexpectedly($?). Agent exiting.");
    }
}

##******************************************************************************
## Method _SpawnSnort private
## Description  : Spawn the Snort process
## Parameters   : none
## Return value : none
##******************************************************************************
sub _SpawnSnort {
    my($self) = shift;
    my($agent) = $self->{SnortAgent};

    my($Init) = sub {

        my($LogRotate) = sub {
            my($filename) = @_;
            my($sfx) = '0000';
            my($newname) = $filename;

            while(-e $newname) {
                $newname = "$filename." . ++$sfx;
            }

            rename($filename, $newname);
        };


            exists($agent->{PortScanFile})
        and &$LogRotate($agent->{PortScanFile}{name});

        &$LogRotate($agent->{AlertFile}{name});

        $agent->{seek} = 0;
    };

    my($Config) = sub {
        local *FIC;

        # Create configuration file
        #
            open(FIC, ">$agent->{ConfigFile}{name}")
        or  Die("Cannot open '$agent->{ConfigFile}{name}' for writing: $!");
        print(FIC $agent->{ConfigFile}{text});
        close(FIC);
        delete($self->{SnortAgent}{ConfigFile});
    };

    my($RunSnort) = sub {
        my($pipe) = new IO::Pipe;
        my($pid);
        my($exec) = $agent->{Exec};
        my($buf) = '';
        my($rin) = '';
        my($output) = '';

        $pid = fork();

            $pid < 0
        and die "Cannot fork(): $!";

            $pid
        or  do {
            local $SIG{PIPE} = 'IGNORE';
            $pipe->writer();
            open(STDOUT,">/dev/null");
            open(STDERR,">&".fileno($pipe));
            exec(@$exec);
        };

        $pipe->reader();
        $pipe->blocking(0);


        vec($rin, fileno($pipe), 1) = 1;

        while(1) {
                select($rin, undef, undef, 3)
            or  Die("Error running Snort:\ncmd = " .
                    join(' ', @$exec) .  "\n" .
                    ('-' x 60) .  "\n" .
                    $output);

            $pipe->sysread($buf, 255, length($buf));
            
            while($buf =~ s/^([^\n]*?\n)//) {
                my($line) = $1;

                    $line =~ /^\s*-\*>\s+Snort!\s+<\*-\s*$/
                and do {
                    SysLog("snort[$pid] running on interface '$agent->{Interface}'");
                    $snortpid = $pid;
                    close($pipe);
                    return;
                };
                $output .= $line;
            }
        }
    };

    $SIG{CHLD} = sub { ++$death };

    &$Init();
    &$Config();
    &$RunSnort();
}

##******************************************************************************
## Method _ParseConfiguration private
## Description  : Spawn the Snort process
## Parameters   : none
## Return value : none
##******************************************************************************
# $self->{SnortAgent} structure out of _ParseConfiguration:
# 
# {
#     Alarms => { # May not exist
#         List => [ [ 'alarm-name', 'alarm-type' ], ... ],
#       Set  => {
#          SnortDeath => alarm-number, # May not exist
#          Threshold => [ # May not exist. Sorted reverse by threshold-values 
#              [ alarm-number, threshold-value ],
#              ...
#          ]
#       }
#     },
#     Classification => { # Mandatory
#         Tag => { # Mandatory
#           tag-name => priority-level, # At least one
#           ...
#       }
#     },
#     ConfigFile => { # Mandatory
#         name => file-name, # Mandatory
#       text => 'content'  # Mandatory
#     },
#     Exec => [ 'executable-name', 'arg1', 'arg2', ... , 'argn' ],
#     Filters => [ # May not exist
#         { re => regexp,   # Maybe undef if sid is defined
#           sid => [ [ include, exclude ], ... ], # May be undef if re is defined
#         priority  # May be undef
#       ],
#       ...
#     ],
#     Interface => 'device-name', # Mandatory
#     Logs => { # Mandatory
#         path => 'path-name' # Mandatory
#       logpackets => boolean # Mandatory
#     },
#     AlertFile => { # Will be used be _Collection
#         name => 'alert-filename',
#         seek => seek-value
#     }
# }
#
# include and exclude from {Filters}[n]{sid} may be both empty
# otherwise, their structure is:
#
# [ [ value, ...], [ value, ... ], [ value, ...] ]
#
# where value is one of:
#
# id-number or [ id-low-bound-range, id-high-bound-range ]
#
# if any bound-range is undef, that means it is open
#
#

sub _ParseConfiguration {
    my($self) = shift;
    my($file) = @_;
    my($parser, $node);

    my($SnortAgent) = sub {
        my($node) = @_;
        my($agconf) = {};

        my($Executable) = sub {
            my($node) = @_;
            my($bin, $name);

                $node = $node->getElementsByTagName('Executable', 0)->item(0)
            or  Die "$file: missing 'Executable' tag";

                $name = $node->getAttribute('name')
            or  Die("$file: tag 'SnortConfig': missing attribute 'name'");
                -e $name
            or  Die("$file: tag 'SnortConfig': attribute 'name': '$name' not found");
                -r $name
            or  Die("$file: tag 'SnortConfig': attribute 'name': '$name' not readable");
                -x $name
            or  Die("$file: tag 'SnortConfig': attribute 'name': '$name' not name");

            $agconf->{Exec} = [ $name ];
        };

        my($Alarms) = sub {
            my($node) = @_;
            my($anum) = 0;
            my($alarms) = {List => [], Set => {}};
            my($alarmlist) = $alarms->{List};
            my($alarmset) = $alarms->{Set};

                $node = $node->getElementsByTagName('Alarms', 0)->item(0)
            or  return;

            for my $alarm ($node->getElementsByTagName('*', 0)) {
                my($name, $label, $type);

                $name = $alarm->getNodeName();

                    $name =~ /^(?:SnortDeath|Threshold)$/
                or  Die("Unkown tag '$name'");

                    $type = $alarm->getAttribute('type')
                or  Die("$file: tag 'SnortDeath': missing attribute 'type'");

                $type = ucfirst(lc($type));

                    $type =~ /(?:Info|Warning|Error)$/
                or  Die("$file: tag '$name': unknown alarm type '$type'");
                    
                    $label = $alarm->getAttribute('label')
                or  Die("$file: tag '$name': missing attribute 'label'");

                push(@$alarmlist, [ $label, $type ]);

                    $name eq 'Threshold'
                and do {
                    my($value) = $alarm->getAttribute('value');

                        length($value)
                    or  Die("$file: tag 'Threshold': missing attribute 'value'");

                        exists($alarmset->{$name})
                    or  $alarmset->{$name} = [];

                    $alarmset->{$name} = [ sort { $b->[-1] <=> $a->[-1] }
                                           (@{$alarmset->{$name}},
                                            [ $anum++, $value ]) ];
                    next
                };

                $alarmset->{$name} = $anum++; 
            }

                @$alarmlist
            and $agconf->{Alarms} = $alarms;
        };


        my($Interface) = sub {
            my($node) = @_;
            my($dev);

                $node = $node->getElementsByTagName('Interface', 0)->item(0)
            or  Die("$file: Missing 'Interface' tag");
                $dev = $node->getAttribute('dev')
            or  Die("$file: tag 'Interface': missing attribute 'dev'");
            $agconf->{Interface} = $dev;
        };

        my($Network) = sub {
            my($node) = @_;
            my($netaddr) = {
                internal => '$' . $agconf->{Interface} . '_ADDRESS',
                external => '!$INTERNAL'
            };

                $node = $node->getElementsByTagName('Networks', 0)->item(0)
            and do {
                my($internal, $external);

                    $internal = $node->getAttribute('internal')
                and $netaddr->{internal} = $internal;
                    $external = $node->getAttribute('external')
                and $netaddr->{external} = $external;
            };

            $agconf->{Network} = $netaddr;
        };

        my($Variables) = sub {
            my($node) = @_;
            my($vars) = []; 

                $node = $node->getElementsByTagName('Variables', 0)->item(0)
            and do {
                for my $var ($node->getElementsByTagName('Variable', 0)) {
                    my($name, $value);

                        $name = $var->getAttribute('name')
                    or  Die ("$file: tag Variable: missing attribute 'name'");
                        $value = $var->getAttribute('value')
                    or  Die ("$file: tag Variable: missing attribute 'value'");

                push(@$vars, [ $name, $value ]);

                }
                $agconf->{Variables} = $vars
            };
        };

        my($Classification) = sub {
            my($node) = @_;
            my($classification) = { List => [], Label => {}, Tag => {} };
            my($list) = $classification->{List};
            my($labels) = $classification->{Label};
            my($tags) = $classification->{Tag};
            my($tnum)=0;

                $node = $node->getElementsByTagName('Classification', 0)->item(0)
            or  Die("$file: Missing 'Classification' tag");

            for my $level ($node->getElementsByTagName('Level', 0)) {
                my($label, $tag);

                    $label = $level->getAttribute('label')
                or  Die ("$file: tag Level: missing attribute 'label'");
                    $tag = $level->getAttribute('tag')
                or  Die ("$file: tag Level: missing attribute 'tag'");

                $tags->{$tag} = $tnum++;
                $labels->{$label} = $tag;
                push(@$list, [ $label, $tag ]);
            }

                @$list
            or  Die("$file: tag 'Classification': no tags 'Level' found");

            $agconf->{Classification}=$classification;
        };

        my($Filters) = sub {
            my($node) = @_;
            my($filters);

            my($ParseSidFilter) = sub {
                my($input) = @_;
                my($sidfilter) = [];
                my($errmsg) = 'Invalid SID filter';

                my($ParseSidFilterElt) = sub {
                    my($input) = @_;
                    my($includes) = [];
                    my($excludes) = [];

                    my($Value) = sub {
                        $input =~ /\G\s*([0-9]+)?\s*(-)?\s*([0-9]+)?\s*/gc;
                            defined($1)
                        or  defined($2)
                        or  die $errmsg;
                        return defined($2) ? [ $1, $3 ] : $1;
                    };

                        $input =~ /\G\s*!/gc
                    ?   push(@$excludes, &$Value)
                    :   push(@$includes, &$Value());

                    while(pos($input) < length($input)) {
                            $input =~ /\G,/gc
                        or  die $errmsg;

                            $input =~ /\G\s*!/gc
                        ?   push(@$excludes, &$Value)
                        :   push(@$includes, &$Value());
                    }

                    [ $includes, $excludes ]

                };

                for my $tmp (split(/\s*:\s*/, $input)) {
                    push(@$sidfilter, &$ParseSidFilterElt($tmp));
                }

                    @$sidfilter == 3
                or  die $errmsg;

                $sidfilter;
            };

                $node = $node->getElementsByTagName('Filters', 0)->item(0)
            or  return;

            for my $filter ($node->getElementsByTagName('Filter', 0)) {
                my($re, $sid, $class, $prio);
                my($result) = {};

                $re = $filter->getAttribute('pattern');
                $sid = $filter->getAttribute('sid');

                    $re
                or  $sid
                or  Die ("$file: tag Filter: no attribute 'pattern' nor 'sid'");

                $class = $filter->getAttribute('classification') || undef;
                $prio=$filter->getAttribute('priority') || undef;

                
                    $prio = $filter->getAttribute('priority')
                and $result->{prio} = $prio;
                
                    $class
                and do {
                        exists($agconf->{Classification}{Label}{$class})
                    or  Die ("$file: tag Filter: unknown classification '$class'");
                    $class = $agconf->{Classification}{Label}{$class};
                    $result->{class} = $class;

                        $prio
                    or  $result->{prio} = $agconf->{Classification}{Tag}{$class};
                };


                    $re
                and do {
                    $re = qr/\[\*\*\]\s*\[[0-9]+:[0-9]+:[0-9]+\].*(?:$re).*\[\*\*\]/;
                    $result->{re} = $re;
                };
                
                    $sid
                and do {
                    my($sidfilter);

                    eval { $sidfilter = &$ParseSidFilter($sid) };

                        $@
                    and Die("$file: tag Filter: sid expr '$sid': $@"); 

                    $result->{sid} = $sidfilter;
                };

                push(@$filters, $result);

            }

            delete($agconf->{Classification}{Label});
            $agconf->{Filters}=$filters;
        };

        my($Logs) = sub {
            my($node) = @_;
            my($path);
            my($logpackets) = 0;


                $node = $node->getElementsByTagName('Logs', 0)->item(0)
            or  Die("$file: Missing 'Logs' tag");

                $path = $node->getAttribute('path')
            or  Die("$file: tag 'Logs': attribute 'path': missing value");

                -e $path
            or  Die("$file: tag 'Logs': attribute 'path': '$path' not found");
                -d $path
            or  Die("$file: tag 'Logs': attribute 'path': '$path' not a directory");
                -r $path
            or  Die("$file: tag 'Logs': attribute 'path': '$path' not readable");
                -w $path
            or  Die("$file: tag 'Logs': attribute 'path': '$path' not writable");

            $logpackets = $node->getAttribute('logpackets') ? 1 : 0;

            $agconf->{Logs} = { path => $path, logpackets => $logpackets };
            $agconf->{ConfigFile}{name} = "$path/tmp.conf";
            $agconf->{AlertFile}{name} = "$path/alert";
            $agconf->{AlertFile}{seek} = 0;
        };

        my($Preprocessors) = sub {
            my($node) = @_;
            my($list) = [];

                $node = $node->getElementsByTagName('Preprocessors', 0)->item(0)
            and do {
                push(@$list,
                 map { { name => $_->[0], options => $_->[1] } }
                     map {
                             $_->getAttribute('name')
                         or  Die ("$file: tag Processor: missing attribute 'name'");

                         [ $_->getAttribute('name'), $_->getAttribute('options') ]
                     } $node->getElementsByTagName('Processor', 0));

                $agconf->{Preprocessors} = $list;
            };
        };

        my($Rules) = sub {
            my($node) = @_;
            my($files) = [];
            my($order) = [];
            my($path, $tmp);

                $node = $node->getElementsByTagName('Rules', 0)->item(0)
            or  Die("$file: Missing 'Rules' tag");

                $path = $node->getAttribute('path')
            or  Die("$file: tag 'Rules': attribute 'path': missing value");

                -e $path
            or  Die("$file: tag 'Rules': attribute 'path': '$path' not found");
                -d $path
            or  Die("$file: tag 'Rules': attribute 'path': '$path' not a directory");
                -r $path
            or  Die("$file: tag 'Rules': attribute 'path': '$path' not readable");

                $tmp = $node->getAttribute('order')
            and do {
                my($hash) = { pass => 0, activation => 0, dynamic => 0, alert => 0, log => 0 };

                push(@$order, map {
                                  my($act) = lc($_);

                                      exists($hash->{$act})
                                  or  Die("$file: tag 'Rules': attribute 'order': invalid action '$act'");

                                      $hash->{$act}
                                  and Die("$file: tag 'Rules': attribute 'order': action '$act' duplicated");

                                  ++$hash->{$act};

                                  $act

                              } split(/\s*,\s*/, $tmp));

                    @$order == keys(%$hash)
                or  Die(  "$file: tag 'Rules': attribute 'order': missing action(s) : '"
                        . join("', '", grep { $hash->{$_} == 0 } keys(%$hash))
                        . "'");
            };

            push(@$files,
                 map {
                     my($name, $fullname);

                         $name = $_->getAttribute('file')
                     or  Die ("$file: tag RulesFile: missing attribute 'file'");

                     $fullname = "$path/$name";

                         -e $fullname
                     or  Die ("$file: tag RulesFile: attribute 'name': '$fullname' file not found");
                         -r $fullname
                     or  Die ("$file: tag RulesFile: attribute 'name': '$fullname' file not readable");

                     $name
                 } $node->getElementsByTagName('RulesFile', 0));

                @$files
            or  Die("$file: tag 'Rules': missing 'RulesFile' tag");

            $agconf->{Rules}{Path} = $path;
            $agconf->{Rules}{Files} = $files;
                @$order
            and $agconf->{Rules}{Order} = $order;
        };

        my($Switches) = sub {
            my($exec) = $agconf->{Exec};

            push(@$exec, '-Afast', $agconf->{Logs}{logpackets} ?  "-b" :  "-N", "-i$agconf->{Interface}",
                         "-c$agconf->{ConfigFile}{name}", "-l$agconf->{Logs}{path}" );
        };

        my($ConfText) = sub {
            my($text) = "\n";
            my($class, $path);

            $text .= "var INTERNAL $agconf->{Network}{internal}\n";
            $text .= "var EXTERNAL $agconf->{Network}{external}\n";

            $text .= "\n";

            for my $var (@{$agconf->{Variables}}) {
                $text .= "var $var->[0] $var->[1]\n";
            }

            $text .= "\n";

            for my $pp (@{$agconf->{Preprocessors}}) {
                $text .= "preprocessor $pp->{name}";
                    $pp->{options}
                and $text .= ": $pp->{options}";
                $text .= "\n";
            }

            $class=$agconf->{Classification}{List};
            for(my $i = 0; $i < @$class; ++$i) {
                $text .= "\nconfig classification: ".
                         join(',',(@{$class->[$i]}, $i));
            }

            $text .= "\n\n";

                exists($agconf->{Rules}{Order})
            and do {
                my($order) = $agconf->{Rules}{Order};

                $text .= "config order: " . join(' ', @$order) . "\n\n"; 
            };

            $path = $agconf->{Rules}{Path};
            for my $rf (@{$agconf->{Rules}{Files}}) {
                $text .= "include $path/$rf\n";
            }

            $agconf->{ConfigFile}{text} = $text;
        };

        my($PortScan) = sub {
            my($list);

                exists($agconf->{Preprocessors})
            or  return;

            $list = $agconf->{Preprocessors};

            for (@$list) {
                my($name) = $_->{name};
                my(@options) = split('\s+', $_->{options});

                    $name eq 'portscan'
                or  next;

                    @options > 3
                or  last;

                $agconf->{PortScanFile}{name} =
                    "$agconf->{Logs}{path}/$options[3]";
            }
        };


            $node = $node->getElementsByTagName('SnortAgent', 0)->item(0)
        or  Die "$file: missing 'SnortAgent' tag";

        &$Executable($node);
        &$Alarms($node);
        &$Interface($node);
        &$Network($node);
        &$Variables($node);
        &$Classification($node);
        &$Filters($node);
        &$Logs($node);
        &$Preprocessors($node);
        &$Rules($node);
        &$Switches();
        &$ConfText();
        &$PortScan();

        delete($agconf->{Classification}{List});
        delete($agconf->{Network});
        delete($agconf->{Preprocessors});
        delete($agconf->{Rules});
        delete($agconf->{Variables});

        $self->{SnortAgent} = $agconf;

    };

    my($LoadSpecificNode) = sub {
        my($parser, $document, $rootnode);

        # Create a XML::DOM::parser object
        #
            $parser = new XML::DOM::Parser()
        or  Die("XML::DOM::Parser: $!");

        # Parse and get the root document node
        #
            $document = $parser->parsefile( $file )
        or  Die("$file :$!");

        # Get Configuration tag node
        #
            $rootnode = $document->getElementsByTagName('Configuration', 0)->item(0)
        or  Die("$file: tag 'Configuration' not found");

        # And then the Specific tag node
        #
            $node = $rootnode->getElementsByTagName('Specific', 0)->item(0)
        or  Die("$file: tag 'Specific' not found");
    };

    &$LoadSpecificNode();
    &$SnortAgent($node);

}

##******************************************************************************
## Method _Collection private
## Description  : Collect datum and alerts from alert files
## Parameters   : none
## Return value : none
##******************************************************************************
sub _Collection {
    my($self) = shift;
    my($agconf) = $self->{SnortAgent};
    my($tags) = $agconf->{Classification}{Tag};
    my($alert) = $agconf->{AlertFile}{name}; 
    my($seek) = $agconf->{AlertFile}{seek}; 
    my($collection) = [];
    my($alarms) = [];
    my($size, $tag, $prio);
    local *FIC;

    my($ProcessLine) = sub {
        my($line) = @_;
        
        my($AddStat) = sub {
            my($tag, $prio) = @_;

                defined($tag)
            and do {

                    exists($tags->{$tag})
                or  do {
                    SysLog("unknown classification '$tag'");
                    return
                };

                ++$collection->[$tags->{$tag}];

            };

                defined($prio)
            and exists($agconf->{Alarms})
            and exists($agconf->{Alarms}{Set}{Threshold})
            and do {
                my($thresholds) = $agconf->{Alarms}{Set}{Threshold};
                for my $threshold (@$thresholds) {
                    my($number, $value) = @$threshold;

                        $prio >= $value
                    or  next;

                    ++$alarms->[$number];
                    last
                }
            };
        };

        my($FilterMatch) = sub {
            my($line, $filters, $class, $prio) = @_;

            my($SidFilterMatch) = sub {
                my($filter, $sid) = @_;
                my($ok) = 0;

                my($Match) = sub {
                    my($filters, $id) = @_;
                    my($ok);

                    for my $range (@$filters) {
                            ref($range)
                        and do {
                            my($low, $high) = @$range;

                                defined($low)
                            and $id < $low
                            and next;

                                defined($high)
                            and $id > $high
                            and next;

                            return 1;

                        };

                            $id == $range
                        and return 1;
                    }
                };

                for my $slot (0..2) {
                    my($includes, $excludes) = @{$filter->[$slot]};
                    my($id) = $sid->[$slot];

                        &$Match($includes, $id)
                    or  return 0;

                        &$Match($excludes, $id)
                    and return 0;
                }
                return 1;
            };

            for my $filter (@$filters) {
                my($sid);

                    exists($filter->{re})
                and $line !~ $filter->{re}
                and next;

                    exists($filter->{sid})
                or  next;

                    @$sid = $line =~ /\[\*\*\]\s*\[([0-9]+):([0-9]+):([0-9]+)\]/
                or  Die "Cannot find SID in alert line: '$line'";

                    &$SidFilterMatch($filter->{sid}, $sid)
                and do {
                        not exists($filter->{class})
                    and not exists($filter->{prio})
                    and return ();

                        exists($filter->{class})
                    and $class = $filter->{class};
                        exists($filter->{prio})
                    and $prio = $filter->{prio};

                    return($class, $prio);
                };
            }
            return($class, $prio);
        };

        ($tag) = $line =~ /\[Classification:\s*([^]]+?)\s*\]/;
        ($prio) = $line =~ /\[Priority:\s*(\d+)\]/;


            exists($agconf->{Filters})
        and not ( ($tag, $prio) = &$FilterMatch($line, $agconf->{Filters}, $tag, $prio) )
        and return; # Match but drop the line

            $tag
        or  $prio
        or  do {
            SysLog("alert file:$line: Neither Classification nor Priority found");
            return
        };

        &$AddStat($tag, $prio);
          
    };

    $size = (stat($alert))[7] || 0;

        ($size > $seek)
    and do {
        my($buf);

            open(FIC, "<$alert")
        or  Die("error opening alert file '$alert': $!");
    
        sysseek(FIC, $seek, 0);

        sysread(FIC, $buf, $size - $seek);
    
        while ($buf =~ /\G([^\n]*)\n/gc) {
            my($line) = $1;
            my($tag);

            $seek += length($line) + 1;
    
            &$ProcessLine($line);
        }

        $agconf->{AlertFile}{seek} = $seek;

        close(FIC);

    };

        $death
    and exists($agconf->{Alarms})
    and exists($agconf->{Alarms}{Set}{SnortDeath})
    and ++$alarms->[$agconf->{Alarms}{Set}{SnortDeath}];

    ($collection, $alarms)

}


1;

__END__

=head1 NAME

SnortAgent.pm - Extension Perl pour LogTrend : Agent Snort

=head1 SYNOPSIS

  use LogTrend::Agent::SnortAgent;

  LogTrend::Agent::SnortAgent->new();
  
=head1 DESCRIPTION

Il s'agit du module d'extension implmentant un agent Snort pour
LogTrend.

Il n'est normalement pas utilis directement, mais via son interface
appelle SnortAgent.

Les diffrents modificateurs des agents LogTrend s'applique pour
celui-ci.

=head1 PRE-REQUIS

Cet agent ncessite une version de snort(1) suprieure ou gale 
la version 1.8beta5 car il utilise ses nouvelles caractristiques
de classification/priorit. La version 1.8 finale est prvue pour
un avenir proche. 

=head1 FONCTIONALITES

L'agent Snort lance une processus snort(1) en tache de fond, les
diffrents paramtres de ce dernier tant dfinis par son fichier
de configuration, par dfaut, F</etc/SnortAgent.conf>.

Il va ensuite,  chaque << collecte >>, extraire du fichier d'alertes
gnrer par snort(1) les informations qu'il remontera au serveur
central.

Ces information sont de deux types :

=over 4

=item *

Des donnes sur le nombre d'vnements dtects pour une classification
donne depuis la dernire collecte. Cette classification est compltement
libre et configurable, mais doit tre arrte avant la premire collecte.

=item *

Des alarmes qui seront dclenche en fonction d'un seuil li au niveau de
priorit attach  la rgle qui a gnr l'alerte. Tout comme pour les
donnes, celle-ci sont libres et configurables, mais doivent avoir t
dcides avant la premire collecte.

=back

Lors de son lancement, si le fichier d'alerte exite dj, celui-ci est
renomm,  des fins d'archivage, sous le nom F<alert.XXX> o XXX est
un numro squentiel complt  gauche par des zros.

=head1 CONFIGURATION

La configuration de l'agent Snort se fait par l'intermdiaire d'un fichier
de configuration en XML.

Comme tous les agents LogTrend, le fichier de configuration se divise en
deux partie, la partie gnrique et la spcifique :

    <?xml version="1.0" standalone="no"?>
    <!DOCTYPE Configuration SYSTEM "Configuration.dtd">
    <Configuration>
       <Generic>
       ....
       </Generic>
       <Specific>
           <SnortAgent>
           ...
           </SnortAgent>
       </Specific>
    </Configuration>
    
La partie gnrique est dcrite dans la documentation du module
LogTrend::Agent.

La partie spcifique est regroupe sous un tag appell C<SnortAgent>, qui
contient les tags suivants :

    <Executable> [Obligatoire] Executable C<snort>.

        Attribut :

            name [Obligatoire]. Le nom complet de l'excutable snort

        Exemple :

            <Executable name="/etc/snort-1.8b4/snort" />

    <Alarms> [Facultatif] Dfinition des alarmes.
        Au moins un tag doit tre dfini.

        Tags :

            <SnortDeath> [0..1] Alerte  la mort prmature du
                processus Snort.

                Attributs :

                    label [Obligatoire] Libell de l'alerte
                    type [Obligatoire] Type d'alarme, parmis
                        "info", "warning", "error"

                <Threshold> [0..n] Alarme de seuil. Seule la plus leve
                    correspondante sera leve lors d'une collecte.

                    Attributs :

                    label [Obligatoire] Libell de l'alerte
                    value [Obligatoire] Valeur de seuil
                    type [Obligatoire] Type d'alarme, parmis
                        "info", "warning", "error"

        Attributs : Aucuns

        Exemple :

            <Alarms>
                <SnortDeath label="Unexpected death of snort process" type="error" />
                <Threshold label="Low-risk traffic" value="1" type="info" />
                <Threshold label="Medium-risk traffic" value="7" type="warning" />
                <Threshold label="High-risk traffic" value="16" type="error" />
            </Alarms>

    <Interface> [Obligatoire]. Interface  surveiller.

        Attributs :

            dev [Obligatoire] Priphrique  surveiller.

        Exemple :

            <Interface dev="eth0" />

    <Networks> [Obligatoire] Dfinition des rseaux locaux et externes.
        Suit les conventions du fichier de configuration de Snort et
        defini deux variables : INTERNAL et EXTERNAL.

        Attribut :

            internal [Obligatoire] Adresse CIDR du/des rseaux locaux
            external [Obligatoire] Adresse CIDR du/des rseaux externes

        Exemple :

            <Networks internal="[192.168.10.0/24,192.168.20.0/24]"
                      external="!$INTERNAL" />

    <Variables> [Facultatif] Permet de dfinir des variables Snort.

        Tags :

            <Variable> [1..n] Dfinition d'une variable Snort

                Attributs :
                
                    name [Obligatoire] Nom de la variable
                    value [Obligatoire] Valeur de la variable

        Attributs : Aucun

        Exemple :

            <Variables>
                <Variable name="HOME_NET" value="$INTERNAL" />
                <Variable name="EXTERNAL_NET" value="$EXTERNAL" />
            </Variables>

    <Classification> [Obligatoire] Dfinition des tags de classification
        associ aux rgle d'alerte utilises. Les priorits commence 
        zro pour le premier tag <Level> et sont incrmentes de un 
        chaque nouveau tag <Level>.

        Tags :

            <Level> [1..n] Dfinition d'un niveau de classification/
                priorit

                Attributs :

                    level [Obligatoire] Cls du niveau pour les
                        rgles d'alerte
                    tags [Obligatoire] Libell du niveau

        Attributs : Aucun.

        Exemple (tir de la classification des << whitehats >>) :

            <Classification>
                <!-- low risk: [0..6] -->
                <Level label="not-suspicious"
                       tag="policy traffic that is not suspicious" />
                <Level label="suspicious"
                       tag="suspicious miscellaneous traffic" />
                <Level label="info-failed"
                       tag="failed information gathering attempt" />
                <Level label="relay-failed"
                       tag="failed relay attempt" />
                <Level label="data-failed"
                       tag="failed data integrity attempt" />
                <Level label="system-failed"
                       tag="failed system integrity attempt" />
                <Level label="client-failed"
                       tag="failed client integrity attempt" />
                <!-- med risk: [7..15] -->

                ...

                <!-- high risk: [16..20] -->

                ...

            </Classification>

    <Filters> [Facultatif] Dfinition de filtrage sur les messages
        d'alerte gnrs par Snort.

        Tags:
        
            <Filter> Dfini un filtre sur le contenu des messages d'alerte.
                Si les attributs "classification" et "priority" sont omis, le
                message est ignor. Ceux prsent en exemple devraient tre
                dfinis car il sont gnrs par les pr-processeur les plus
                couramment utiliss et peuvent gnrer du bruit, surtout
                les messages de "portscan". De plus, pour les preprocesseurs,
                les valeurs de classification/priorit ne sont pas gres :
                c'est, aujourd'hui, l'unique faon de les traiter.
                Le filtrage peut se faire soit sur motif dans le corps du
                message d'alerte, soit par le SID, soit les deux.
                
                Le filtrage par motif utilise une expression rationnelle
                Perl qui doit correspondre dans le texte de l'alerte
                (entre "[**] [x:y,z]" et " [**]" exclus). Pour ajouter
                des switchs, utiliser l'option (?switch) : par exemple
                pour ignorer la casse '(?i)ids283|ids128'.

                Le filtrage par SID est plus complexe. Les SID sont
                composs de trois champs :
                  - Le gnrateur du message (1 pour le moteur de rgles)
                  - Le numro du message que l'on peut trouver dans le
                    fichier generators.h des sources snort (ou bien le
                    numro affect  la rgle si le gnrateur est 1)
                  - Pour les rgles, le numro de version.

                Les dfinitions se font de manire identique pour les trois
                champs, spares par le caractre ':' (les blacs sont ignors
                dans la totalit de la spcification du filtre par SID).

                Les dfinitions peuvent comporter plusieurs valeurs, spares
                par le caractre ','.

                Chaque valeur est soit un nombre, soit un intervalle, et peut
                tre inverse logiquement (exclusion) en la prfixant par le
                caractre '!'.

                Les intervalles se dfinissent par le seuil bas (un nombre),
                qui peut tre omis, auquel cas l'intervalle est ouvert  gauche,
                suivit du caractre '-' et du seuil haut (aussi un nombre), qui
                peut lui aussi tre omis, auquel cas l'intervalle est ouvert 
                droite. Un intervalle peut tout  fait tre ouverts des deux cts,
                auquel cas il se compose de l'unique caractre '-'.

                Exemples:

                        sid="100: 1: -"

                    specifie le preprocesseur 'portscan', message numro 1, quelle
                    que soit la valeur du troisime champ.

                        sid="100: -, !1: -"

                    specifie le preprocesseur 'portscan', tous les messages sauf
                    le numro 1, quelle que soit la valeur du troisime champ.

                        sid="1: 45-60, !48, !50, !55: 1-"

                    spcifie le moteur de rgles, les rgle 45  60, sauf les
                    rgles 48, 50 et 55, pour toutes les version suprieure ou
                    gale  1.

                Pour qu'un fitre corresponde  un message, l'ordre d'valuation
                est le suivant :

                    - Si un motif est prcis, si celui-ci ne correspond pas,
                      le fitre ne correspond pas.
                    - Si un fitrage par SID est prcis, pour chacun des champs:
                        - Pour chaque lment d'inclusion dans l'ordre de leurs
                          spcifications si la valeur du champ est gal au nombre
                          ou fait partie de l'intervalle alors il y a correspondance
                          potentielle et les tests d'exclusion sont appliqus.
                          Si aucune correspondance n'est trouve, le fitre ne
                          correspond pas.
                        - Enfin, pour chaque lment d'exclusion (ceux prcds par
                          le caractre '!') dans l'ordre de leurs spcifications,
                          si la valeur du champ est gale au nombre ou fait partie
                          de l'intervalle, alors le fitre ne correspond pas.
                          

                Attributs [Au moins 1 des attributs "pattern" ou "sid" doit tre
                           prsent] :

                    pattern [Facultatif] Expression rationnelle  recherche
                        dans les message d'alerte
                    sid [Facultatif] Filtre par SID, comme dcrit ci-dessus
                    classification [Facultatif] Si prcis, doit apparaitre comme
                        attribut "label" d'un des tags <Level> du tag <Classification>
                        et y associe le message si "pattern" et/ou sid correspondent.
                    priority [Facultatif] Permet d'imposer un niveau de priorit.
                        

        Attributs : Aucun

        Exemple :
            <Filters>
                <!-- preprocessor SIDs are from file "generators.h" -->
                <!-- 100: portscan -->
                <!-- Log only start of portscan messages -->
                <Filter sid="100:1:-"
                        classification="info-attempt" priority="2"/>
                <!-- Ignore noisy portcan messages -->
                <Filter sid="100:-,!1:-" />
                <!-- 101: minfrag -->
                <Filter sid="101:-:-"
                        classification="suspicious" />
                <!-- 102: http_decode -->
                <Filter sid="102:-:-"
                        classification="system-attempt" />
                <!-- 103: defrag -->
                <Filter sid="103:1:-"
                        classification="denialofservice" />
                <Filter sid="103:2:-"
                        classification="suspicious" />
                <!-- 105: bo -->
                <Filter sid="105:-:-"
                        classification="system-success" />
            </Filters>


    <Preprocessors> [Facultatif] Dfinition des prprocesseurs Snort 
        utiliser. Il est recommend de toujours inclure "defrag" et si possible
        "stream2". Si il n'y a pas de serveur HTTP, "http_decode" peut tre
        omis. "portscan-ognorehost" est chaudement recommand si "portscan" est
        utilis (beaucoup de fausses alertes sinon).

        Tags:

            <Processor> [1..n] Dfinis un prprocesseur a utiliser.

                Attributs :

                    name [Obligatoire] Nom du prprocesseur (voir la documentation
                        de snort(1))
                    options [Facultatif] Options  passer au prprocesseur

        
        Attributs : Aucun

        Exemple :
                    
            <Preprocessors>
                <Processor name="defrag" />
                <Processor
                    name="stream2"
                    options="timeout 23, ports 21 23 25 80 110 143, maxbytes 16384" />
                <Processor
                    name="http_decode"
                    options="80 2301" />
                <Processor
                    name="rpc_decode"
                    options="111 32771" />
                <Processor
                    name="bo"
                    options="-nobrute" />
                <Processor
                    name="telnet_decode" />
                <Processor
                    name="portscan"
                    options="$INTERNAL 5 5 portscan" />
                <Processor
                    name="portscan-ignorehosts"
                    options="$INTERNAL" />
            </Preprocessors>

    <Logs> [Obligatoire] Paramtres des fichiers de trace.

        Attributs :

            path [Obligatoire] Repertoire ou les fichiers de trace doivent
                tre gnrs
            logpackets [Facultatif] "y" ou "n". Indique s'il faut ou non
                enregister les paquets ayant gnr une alarme.
                Hautement suggr pour investigations.

        Exemple :

            <Logs path="/var/log/snort/eth0" logpackets="y" />


    <Rules> [Obligatoire] Dfinitions des fichiers de rgle d'alertes.

        Tags :

            <RulesFile> [1..n] Nom du fichier contenant des rgle d'alerte.

                Attributs :

                    file [Obligatoire] nom du fichier de rgles

        Attributs :

            path [Obligatoire] Chemin d'accs aux fichiers de rgles
            order [Facultatif] Ordre d'application des rgles.
                Liste spare par des ',' des 5 mots cls :
                    - pass
                    - activation
                    - dynamic
                    - alert
                    - log
                Le dfaut est celui de Snort :
                    activation, dynamic, alert, pass, log


        Exemple :

            <Rules path="/etc/snort.d">
                <RulesFile file="rules" />
            </Rules>

=head1 COMMENTAIRES/RECOMMENDATIONS

L'agent utilise les nouvelles fonctionnalits de la future ( ce jour,
le 18 juin 2001) version 1.8 de Snort,  savoir le plug-in de
classification et de priorit et l'intgration des options dans le
fichier de configuration.

Deux jeux de rgles de filtrage sont disponibles :

=over 4

=item *

Celles fournies avec la distribution de Snort

=item *

Celles disponibles depuis le site des << whitehats >>, compiles
depuis la base de donne heuristique archNIDS (voir
http://www.whitehats.com/ids/index.html)

=back

Aujourd'hui, la classification faite pour la base archNIDS me parait
la plus aboutie, c'est celle-ci que je conseillerais, toutefois il
ne faut surtout pas ngliger de les mettre  jour rgulirement et
de vrifier si de nouvelles rgles qui pourraient vous tre ncessaires
n'ont pas t publies par ailleurs. Enfin, les rgles et les
classifications livres avec Snort pourront certainement s'amliorer
et se stabiliser d'ici sa sortie en << stable >>.

La dtection d'intrusion n'est pas une science exacte : certaines
rgles vous gnreront ventuellement de fausses alarmes. Plutt que
de les commenter, mieux vaut les analyser finement et au besoin les
modifier ou crer des rgle de type 'pass' spcifique et modifier
l'ordre d'application des rgles (voir plus loin).

Le pr-processeur "defrag" doit tre activ : il a en charge de
dfragmenter les paquets avant de les passer aux filtres, et nombre
d'attaques utilisent la fragmentation pour ne pas tre dtectes par
les NIDS.

Les diffrents pr-processeurs, dans la version actuelle, ne grent
pas les classifications ni les priorit Ils retournent ventuellement
des valeur qui n'ont aucune signification. De plus, le pr-processeur
"portscan" gnre beaucoup trop de bruit. D'o l'utilit des filtres
sur les messages d'alerte que je conseille de conserver en l'tat.

L'option logpackets enregistre sous forme binaire (peu de CPU) les
paquets ayant gnr un alerte. Ces fichier peuvent tre relu par
Snort ou Tcpdump ultrieurement. Cet enregistrement est indispensable
pour pouvoir analyser les causes d'un alerte. Je ne peux que recommender
fortement de la mettre  "y".

L'attribut 'order' du tag <Rules> permet de spcifier l'ordre dans lequel
seront traites les rgles. Ceci ne concerne absolument pas les pr-processeurs
dont le travail est fait en amont. Si des rgles de type 'pass' sont dfinies,
il est important de r-ordonner l'ordre d'application car, par dfaut,
les 'alert' sont trait AVANT les 'pass' qui sont eux-mmes AVANT les 'log'.
Dans ce cas, l'ordre correct est : activate, dynamic, pass, alert, log.
Mais attention, si vous crez une rgle 'pass UDP any any -> any any'
(je caricature) et que vous mettez l'odre ci-dessus, vous ne traiterez plus
AUCUN paquets UDP.

Enfin, pour ajouter des rgles propres au site local, je conseille de
constituer un fichier spar, appell 'local.rules' et de le dclarer
dans la configuration.


=head1 AUTHOR

Franois Dsarmnien -- Atrid Systmes SAS (f.desarmenien@atrid.fr)

=head1 COPYRIGHT

Copyright 2001, Atrid Systme SAS.

Distribu selon les mmes terme le projet.

Licensed under the same terms as LogTrend project is.

=head1 WARRANTY

THIS SOFTWARE COMES WITH ABSOLUTLY NO WARRANTY OF ANY KIND.
IT IS PROVIDED "AS IS" FOR THE SOLE PURPOSE OF EVENTUALLY
BEEING USEFULL FOR SOME PEOPLE, BUT ONLY AT THEIR OWN RISK.


=cut


